#include "cthugha.h"
#include "sound.h"
#include "options.h"
#include "display.h"
#include "cd_player.h"
#include "information.h"
#include "interface.h"
#include "server.h"

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <math.h>
#include <stdio.h>
#include <sys/ioctl.h>
#include <stdlib.h>
#include <sys/time.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <signal.h>
#include <memory.h>

/*
 * public Variables 
 */
char2 * sound_data;			/* Sound-Data in 8 bit stereo */
char2 * sound_data_shared;		/* pointer to the shared memory */
int sound_bsize = 0;			/* size of sound-buffer (samples) 
					   alway a power of 2 */
int sound_fork = 0;			/* read in an extra process */
int sound_use[snddev_max] = {1,0,1,0,0,0};

/* 
 * local variables
 */
static int sound_child;
static int sound_shm_key;
static int sound_alloc = 0;		/* sound currently allocated */

int sound_use_clear() {
    memset(sound_use, 0, sizeof(int)*snddev_max);
    return 0;
}

/* 
 *  Initialize the Sound-Interface
 */
int init_sound() {

    /* make sue the sound_volumes is initialized */
    mixer_init_volumes();

    if ( sound_use[snddev_cd] )
	init_cd();

    sound_bsize =  1 << log2(max(BUFF_WIDTH, BUFF_HEIGHT));

    if( alloc_sound() )
	return 1;

    return 0;
}


/*
 * Clean up the sound-interface
 */
int exit_sound() {
    if ( sound_use[snddev_cd] )
	exit_cd();

    return free_sound();
}

/* 
 * get the sound devices
 */
int alloc_sound() {
    /* check, if already allocated */
    if(sound_alloc)				
	return 0;

    /* 
     * initialize only the sound devices needed
     */

    if( sound_use[snddev_sock]) {
	printfv(2, "  initializing network sound.\n");
	if( sound_net_init() )
	    return 1;
    }

    if( sound_use[snddev_play]) {
	printfv(2, "  initializing play sound.\n");
	if( sound_play_init() )
	    return 1;
    }

    if( sound_use[snddev_mixer]) {
	printfv(2, "  initializing mixer.\n");
	if( sound_set_mixer() )
	    return 1;
    }

    if( sound_use[snddev_dsp_r]) {
	printfv(2, "  initiliazing dsp for reading.\n");
	if( sound_dsp_init(O_RDONLY) )
	    return 1;
    }

    if( sound_use[snddev_dsp_w]) {
	printfv(2, "  initiliazing dsp for writing.\n");
	if( sound_dsp_init(O_WRONLY) )
	    return 1;
    }

    if(sound_fork) 
	sound_fork_process();
    sound_data = cth_memory(NULL, sound_bsize*2,
			    "Can not allocate sound_data.");

    sound_alloc = 1;

    return 0;
}

/*
 * fork the sound reading process
 */
int sound_fork_process() {
    printfv(3,"    starting sound reading process\n");
    if( (sound_shm_key = shmget(IPC_PRIVATE, sound_bsize*2, 
				IPC_CREAT | 0777) ) == -1) {
	printfee("Can not create shared memory segment.");
	sound_fork = 0;
	return 0;
    }
    if( (sound_data_shared=(char2*)shmat(sound_shm_key, 0, 0)) == (void*)-1) {
	printfee("Can not attach shared memory segment.");
	sound_fork = 0;
	return 0;
    }
    switch( sound_child = fork() ) {
    case -1:		/* error */
	printfee("Can not fork sound child.");
	shmdt((char*)sound_data);
	sound_fork = 0;
	break;
    case 0:		/* child */
	sound_data = sound_data_shared;
	do {
	    sound_read();
	} while(1);
    }
    return 0;
}

/*
 * kill the sound reading process
 */
int sound_kill_process() {
    printfv(3,"    killing sound reading process.\n");
    kill(sound_child, 9);		/* very brutal, but efficient */
    shmdt((char*)sound_data_shared);
    shmctl(sound_shm_key, IPC_RMID, 0);

    return 0;
}

/*
 * free the sound device
 */
int free_sound() {

    if(!sound_alloc)
	return 0;

    if(sound_fork) {
	sound_kill_process();
    } else {
	free(sound_data);
    }

    if(sound_use[snddev_sock])
	sound_net_exit();
    if(sound_use[snddev_play])
	sound_play_exit();
    if(sound_use[snddev_dsp_r])
	sound_dsp_exit();

    sound_alloc = 0;
    return 0;
}    

/*****************************************************************************/

int sound_read() {

    if(sound_use[snddev_debug])
	sound_random_read();
    if(sound_use[snddev_sock])
	sound_net_read();
    if(sound_use[snddev_play])
	sound_play_read();
    if(sound_use[snddev_dsp_r])
	sound_dsp_read();
    
    return 0;
}

/*
 * Get a "line" of sound from the soundcard.
 */
int get_sound() {
    if(sound_fork) {
	memcpy(sound_data, sound_data_shared, sound_bsize*sound_stereo);
	return 0;
    }

    sound_read();
    
    return 0;
}

/*****************************************************************************/


int sound_convert(unsigned char * data, int stereo, int bps) {
    unsigned int  * data_i = (unsigned int*)data;
    int i;

    if( (bps == 8) && (stereo == 2) ) {
	for(i=0; i < sound_bsize; i++) {
	    sound_data[i][1] = (int)(*data ++) - 128;
	    sound_data[i][0] = (int)(*data ++) - 128;
	}
    } else if( (bps == 8) && (stereo == 1) ) {
	for(i=0; i < sound_bsize; i++) {
	    sound_data[i][0] = (int)(*data) - 128;
	    sound_data[i][1] = (int)(*data ++) - 128;
	}	
    } else if( (bps == 16) && (stereo == 2) ) {
	for(i=0; i < sound_bsize; i++) {
	    sound_data[i][1] = (int)((*data_i ++) >> 8) - 128;
	    sound_data[i][0] = (int)((*data_i ++) >> 8) - 128;
	}
    } else if( (bps == 16) && (stereo == 1) ) {
	for(i=0; i < sound_bsize; i++) {
	    sound_data[i][1] = (int)((*data_i) >> 8) - 128;
	    sound_data[i][0] = (int)((*data_i ++) >> 8) - 128;
	}
    } 
    return 0;
}


